<?xml version='1.0'?>														 
<!-- 
   - The contents of this file are subject to the Mozilla Public
   - License Version 1.1 (the "License"); you may not use this file
   - except in compliance with the License. You may obtain a copy of
   - the License at http://www.mozilla.org/MPL/
   - 
   - Software distributed under the License is distributed on an "AS
   - IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   - implied. See the License for the specific language governing
   - rights and limitations under the License.
   - 
   - The Original Code is XML Schema Compiler (Pass2).
   - 
   - The Initial Developer of the Original Code is AEA Technology - Engineering Software.
   -  Portions created by AEA Technology - Engineering Software are
   - Copyright (C) 2000 AEA Technology - Engineering Software.  All
   - Rights Reserved.
   - 
   - Contributor(s): Curt Arnold <curta@hyprotech.com>
  --> 
												 
<xsl:transform xmlns:xsl='http://www.w3.org/1999/XSL/Transform' 
     version="1.0"
	 xmlns:xsc="http://XSDComp.sourceforge.net/Namespaces/2000/XMLSchema-compiled">

<xsl:output method="xml" version="1.0" indent="yes"/>

<xsl:key name="simple-types" match="//xsc:simpleType" use="@id"/>
<xsl:key name="simple-types-by-base" match="//xsc:simpleType" use="@base"/>
<xsl:key name="attribute-by-type" match="//xsc:attribute" use="@type"/>
<xsl:key name="attributes" match="//xsc:attribute" use="@id"/>
<xsl:key name="element-by-type" match="//xsc:element" use="@type"/>
<xsl:key name="elements" match="//xsc:element" use="@id"/>
<xsl:key name="equiv-classes" match="//xsc:element[@equivClass]" use="@equivClass"/>

<xsl:key name="attribute-groups" match="//xsc:attributeGroup" use="@id"/>

<!-- matches document root and invokes templates for all child nodes -->
<xsl:template match="/">
	<xsl:apply-templates/>
</xsl:template>

<!--  unrecognized elements and attributes get dropped  -->
<xsl:template match="*">
	<xsl:element name="{name()}">
		<xsl:apply-templates select="@*|text()|*"/>
	</xsl:element>
</xsl:template>

<xsl:template match="@*">
	<xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute>
</xsl:template>

<xsl:template match="text()">
	<xsl:value-of select="text()"/>
</xsl:template>

<xsl:template match="xsc:attribute">
	<xsc:attribute>
		<xsl:apply-templates select="@id"/>
		<xsl:apply-templates select="@namespace"/>
		<xsl:apply-templates select="@name"/>
		<xsl:apply-templates select="@global"/>
	</xsc:attribute>
</xsl:template>
	

<xsl:template match="xsc:simpleType">
	<xsl:param name="parent-types"/>

	<!--  check for cyclic inheritance   -->
	<xsl:if test="contains($parent-types,concat(@id,' '))">
		<xsl:message terminate="yes">Fatal Error: simpleType id="<xsl:value-of select="@id"/>", name="<xsl:value-of select="@name"/> participates in a cyclic inheritance hierarchy.</xsl:message>
	</xsl:if>

	<xsc:simpleType id="{@id}">
		<xsl:apply-templates select="@name"/>
		<xsl:apply-templates select="@namespace"/>

		<!--  reproduce all annotation and facets    -->
		<xsl:apply-templates/>

		<!--  list all types that are lists of this type   -->
		<xsl:variable name="derivedTypes" select="key('simple-types-by-base',@id)"/>

		<xsl:variable name="lists" select="$derivedTypes[@list='1' and @base!=@id]"/>
		<xsl:if test="$lists">
			<xsc:listTypes>
				<xsl:apply-templates select="$lists">
					<xsl:with-param name="parent-types" select="concat(concat($parent-types,' '),@id)"/>
				</xsl:apply-templates>
			</xsc:listTypes>
		</xsl:if>

		<!--  list all types that are derived from this type    -->
		<xsl:variable name="derivedNonLists" select="$derivedTypes[not(@list='1') and @base!=@id]"/>
		<xsl:if test="$derivedNonLists">
			<xsc:derivedTypes>
				<xsl:apply-templates select="$derivedNonLists">
					<xsl:with-param name="parent-types" select="concat(concat($parent-types,' '),@id)"/>
				</xsl:apply-templates>
			</xsc:derivedTypes>
		</xsl:if>

		<!--   okay now all attributes that have this simpleType    -->
		<xsl:apply-templates select="key('attribute-by-type',@id)"/>

		<!--   okay now all elements that have this simpleType    -->
		<xsl:apply-templates select="key('element-by-type',@id)"/>
	</xsc:simpleType>
</xsl:template>

<xsl:template name="compact-namespaces">
	<xsl:variable name="included" select="xsc:namespace"/>
	<xsl:variable name="excluded" select="xsc:namespaceNor/xsc:namespace"/>

	<!-- copy any attribute   -->
	<xsl:apply-templates select="@*"/>

	<!-- express the included namespaces as an IDREFS attribute  -->
	<xsl:if test="$included">
		<xsl:attribute name="included"><xsl:value-of select="$included[position()=1]"/><xsl:for-each select="$included[position()!=1]"><xsl:text> </xsl:text><xsl:value-of select="@namespace"/></xsl:for-each></xsl:attribute>
	</xsl:if>

	<!-- express the excluded namespaces as an IDREFS attribute  -->
	<xsl:if test="$excluded">
		<xsl:attribute name="excluded"><xsl:value-of select="$excluded[position()=1]"/><xsl:for-each select="$excluded[position()!=1]"><xsl:text> </xsl:text><xsl:value-of select="@namespace"/></xsl:for-each></xsl:attribute>
	</xsl:if>
</xsl:template>

<xsl:template match="xsc:anyAttribute">
	<xsc:anyAttribute>
		<xsl:call-template name="compact-namespaces"/>
	</xsc:anyAttribute>
</xsl:template>

<xsl:template match="xsc:any">
	<xsc:anyAttribute>
		<xsl:call-template name="compact-namespaces"/>
	</xsc:anyAttribute>
</xsl:template>

<xsl:template name="expand-attribute">
	<xsl:param name="restrict-top"/>
	<xsl:param name="call-stack"/>
	<xsl:param name="attribute"/>
	<xsl:param name="use"/>
	<xsl:param name="use-value"/>
	<xsl:param name="default"/>
	<xsl:param name="fixed"/>
	<xsl:param name="fixed-value"/>

	<xsl:variable name="match" select="$restrict-top/xsc:attributeUseRestriction[@name=$attribute/@name and @namespace=$attribute/@namespace and @form=$attribute/@form]"/>
	<xsl:message>Found match for <xsl:value-of select="$attribute/@name"/>=<xsl:value-of select="boolean($match)"/></xsl:message>

	<!--  do recursion first to make sure anything we set here is done last   -->
	<xsl:for-each select="$restrict-top/xsc:attributeGroupRef">
		<xsl:if test="contains($call-stack,concat(@id,' '))">
			<xsl:call-template name="expand-attribute">
				<xsl:with-param name="restrict-top" select="."/>
				<xsl:with-param name="call-stack" select="concat(concat(@id,' '),$call-stack)"/>
				<xsl:with-param name="attribute" select="$attribute"/>
				<xsl:with-param name="use" select="$use or $match/@use"/>
				<xsl:with-param name="use-value" select="concat($use,$match/@use)"/>
				<xsl:with-param name="default" select="$default or $match/@default"/>
				<xsl:with-param name="fixed" select="$fixed or $match/@fixed"/>
			</xsl:call-template>
		</xsl:if>
	</xsl:for-each>


	<!--  if we have a match   -->
	<xsl:if test="$match">

		<!-- if this restriction has set a use    -->
		<xsl:if test="$match/@use">

			<!--  lets see if use had been set closer to the top   -->
			<xsl:choose>
				<xsl:when test="$use">
					<!--  can't do the test if we have concatenated multiple levels together  
					            however each level should have been done separately   -->
					<xsl:if test="$use-value='optional' or $use-value='required' or $use-value='prohibited'">
						<xsl:choose>
							<!--  anything goes when lower constraint was optional  -->
							<xsl:when test="$match/@use='optional'"/>

							<!--   oops, they have to be the same  -->
							<xsl:when test="$match/@use!=$use-value">
								<xsl:message terminate="yes">Fatal Error: Restriction of attribute <xsl:value-of select="$match/@name"/> set use=<xsl:value-of select="$use-value"/> when previous restriction had set it to <xsl:value-of select="$match/@use"/>.</xsl:message>
							</xsl:when>
						</xsl:choose>

						<!--   now have to compare it to the attribute itself   -->
						<xsl:choose>
							<xsl:when test="$attribute/@use='optional'"/>
							<!--   oops, they have to be the same  -->
							<xsl:when test="$attribute/@use!=$use-value">
								<xsl:message terminate="yes">Fatal Error: Restriction of attribute <xsl:value-of select="$match/@name"/> set use=<xsl:value-of select="$use-value"/> when attribute definition had set it to <xsl:value-of select="$attribute/@use"/>.</xsl:message>
							</xsl:when>
						</xsl:choose>
					</xsl:if>
				</xsl:when>

				<!--   not set at higher level, go ahead and express it   -->
				<xsl:otherwise>
					<xsl:attribute name="use"><xsl:value-of select="$match/@use"/></xsl:attribute>
				</xsl:otherwise>
			</xsl:choose>
		</xsl:if>
	</xsl:if>
</xsl:template>

								
							

<!--  this expand the immediate children of an attribute group
          pumping them through the restrictions and
		  and then calling any nested attribute groups   -->
<xsl:template name="expand-attribute-level">
	<xsl:param name="restrict-top"/>
	<xsl:param name="call-stack"/>

	<xsl:comment>Entering expand-attribute-level id=<xsl:value-of select="@id"/>, attrUses=<xsl:value-of select="count(xsc:attributeUse)"/>, attrGroups=<xsl:value-of select="count(xsc:attributeGroupRef)"/>.</xsl:comment>

	<xsl:choose>
		<!--  if we have been here before    -->
		<xsl:when test="contains($call-stack,concat(@id,' '))">
			<xsl:message>Warning: Attribute group with id <xsl:value-of select="@id"/> participates in infinite regression.</xsl:message>
		</xsl:when>

		<xsl:otherwise>
			<xsl:for-each select="xsc:attributeUse">
				<xsc:attributeUse>
					<!-- express base values in case we don't have matching restriction   -->
					<xsl:apply-templates select="@attribute"/>
					<xsl:apply-templates select="@use"/>
					<xsl:apply-templates select="@fixed"/>
					<xsl:apply-templates select="@default"/> 

					<xsl:call-template name="expand-attribute">
						<xsl:with-param name="restrict-top" select="$restrict-top"/>
						<xsl:with-param name="attribute" select="key('attributes',@attribute)"/>
					</xsl:call-template>
				</xsc:attributeUse>
			</xsl:for-each>

			<!-- for any nested attribute group references   -->
			<xsl:variable name="new-call-stack" select="concat($call-stack,concat(@id,' '))"/>

			<xsl:for-each select="xsc:attributeGroupRef">
				<xsl:comment>Recursing into <xsl:value-of select="@ref"/></xsl:comment>
				<!--  find the corresponding attribute group   -->
				<xsl:for-each select="key('attribute-groups',@ref)">
					<xsl:comment>Located attributeGroup <xsl:value-of select="@id"/></xsl:comment>
					<!--   and call this routine recursively   -->
					<xsl:call-template name="expand-attribute-level">
						<xsl:with-param name="restrict-top" select="$restrict-top"/>
						<xsl:with-param name="call-stack" select="$new-call-stack"/>
					</xsl:call-template>
				</xsl:for-each>
			</xsl:for-each>
		</xsl:otherwise>
	</xsl:choose>
</xsl:template>


<xsl:template match="xsc:attributeGroup">
	<xsc:attributeGroup>
		<xsl:apply-templates select="@*|xsc:annotation"/>

	
		<xsl:choose>
			<!--  if I have restrictions, I have to pump every attribute
			             through a potential maze of restrictions    -->
			<xsl:when test="xsc:attributeUseRestriction">
				<xsl:comment>Trying to expand restrictions</xsl:comment>
				<xsl:call-template name="expand-attribute-level">
					<xsl:with-param name="restrict-top" select="."/>
				</xsl:call-template>
			</xsl:when>

			<!--   no restrictions, just apply templates on the non-annotation elements   -->
			<xsl:otherwise>
				<xsl:apply-templates select="*[name()!='xsc:annotation']"/>
			</xsl:otherwise>
		</xsl:choose>
	</xsc:attributeGroup>
</xsl:template>

<xsl:template match="xsc:attributeGroupRef">
	<xsc:attributeGroupRef>
		<xsl:apply-templates select="@*|*"/>
	</xsc:attributeGroupRef>
</xsl:template>

<!--  expanding element that are members of equivalance classes   -->
<xsl:template match="xsc:element" mode="element-use">
	<xsl:choose>
		<!--  if it is an extemplar of its own equivClass  -->
		<!-- xsl:when test="key('equiv-classes',@id)" -->
		<xsl:when test="//xsc:element[@equivClass=current()/@id]">
			<xsc:groupUse group="{@id}_equiv_class" minOccurs="1" maxOccurs="1" unbounded="false"/>
		</xsl:when>

		<xsl:otherwise>
			<xsc:elementUse element="{@id}" minOccurs="1" maxOccurs="1" unbounded="false"/>
		</xsl:otherwise>
	</xsl:choose>
</xsl:template>


<xsl:template match="xsc:package">
	<xsc:package>
		<!--  duplicate existing namespace entries   -->
		<xsl:apply-templates select="xsc:namespace"/>

		<!--  test to see if there are any simple types who base is not a simple type   -->
		<xsl:variable name="orphaned-simple-types" select="descendant::xsc:simpleType[@base and not(key('simple-types',@base))]"/>
		<xsl:if test="$orphaned-simple-types">
			<xsl:message terminate="yes">Internal Error: the following simpleTypes had a base attribute that was not a simple type reference, ids=<xsl:for-each select="$orphaned-simple-types"><xsl:value-of select="@id"/>,</xsl:for-each>.</xsl:message>
		</xsl:if>


		<!--  produce tree of simple types but transforming each root simple type  
			  This produces all attributes and all elements that have a simple type  
		-->
		<xsl:apply-templates select="descendant::xsc:simpleType[@id=@base or not(@base)]"/>

		<!-- produce all attribute groups    -->
		<xsl:apply-templates select="descendant::xsc:attributeGroup"/>

		<!--  produce elements    -->
		<xsl:apply-templates select="descendant::xsc:element"/>

		<!--  produce groups corresponding to equivalance classes  -->
		<xsl:for-each select="descendant::xsc:element[@equivClass]">
			<!--   if this is the first appearance of the equivalance class   -->
			<xsl:if test="not(preceding::xsc:element[@equivClass=current()/@equivClass])">
				<xsl:message>Debug: equivClass = <xsl:value-of select="@equivClass"/></xsl:message>
				<!-- xsl:variable name="extemplar" select="key('elements',@equivClass)"/ -->
				<xsl:variable name="extemplar" select="//xsc:element[@id=current()/@equivClass]"/>
				<xsl:choose>
					<xsl:when test="count($extemplar)!=1">
						<xsl:message terminate="yes">Unexpected error: unable to resolve extemplar of equivClass <xsl:value-of select="@equivClass"/> count=<xsl:value-of select="count($extemplar)"/>.</xsl:message>
					</xsl:when>

					<xsl:otherwise>
						<!--  we should do a lot of checks to see if the extemplar is appropriate  -->
						<xsc:group id="{@equivClass}_equiv_class" compositor="choice">
							<xsl:if test="not($extemplar/@abstract='true')">
								<xsc:elementUse element="@equivClass" minOccurs="1" maxOccurs="1" unbounded="false"/>
							</xsl:if>
							<xsl:apply-templates select="/descendant::xsc:element[@equivClass=$extemplar/@id]" mode="element-use"/>
						</xsc:group>
					</xsl:otherwise>
				</xsl:choose>
			</xsl:if>
		</xsl:for-each>

		<!--  produce all groups that have an id    -->
		<xsl:apply-templates select="//xsc:group[@id]"/>

		<xsl:apply-templates select="//xsc:complexType"/>

	</xsc:package>
</xsl:template>

<xsl:template match="xsc:complexType">
	  <!--  we need to check that any derivation doesn't violated the blocks and finals    -->


      <xsc:complexType id="{@id}" group="{@group}" attributes="{@attributes}" name="{@name}" base="{@base}" abstract="{@abstract}">
		<xsl:apply-templates select="xsc:annotation"/>
	  </xsc:complexType>
</xsl:template>

<xsl:template match="xsc:restrictedGroup">
	<xsl:message>Information: element content restriction not currently implemented.</xsl:message>
	<xsc:groupUse group="{@base}" minOccurs="1" maxOccurs="1" unbounded="false"/>
</xsl:template>

<xsl:template match="xsc:element">
	<!--  strip equivalence class     -->
    <xsc:element id="{@id}" name="{@name}" nullable="{@nullable}" namespace="{@namespace}" type="{@type}">
		<xsl:apply-templates select="*"/>
	</xsc:element>
</xsl:template>

<xsl:template match="xsc:annotation">
	<xsc:annotation>
		<xsl:apply-templates select="xsc:appinfo|xsc:documentation"/>
	</xsc:annotation>
</xsl:template>

<xsl:template match="xsc:appinfo">
	<xsc:appinfo>
		<xsl:apply-templates select="@*|*|text()"/>
	</xsc:appinfo>
</xsl:template>

<xsl:template match="xsc:documentation">
	<xsc:documentation>
		<xsl:apply-templates select="@*|*|text()"/>
	</xsc:documentation>
</xsl:template>


</xsl:transform>